-- 	Copyright  2009.	GSI Technology
-- 						Jeff Daugherty
-- 						apps@gsitechnology.com
--  Version: 3.6
--
-- 	FileName: core_burst.vhd / core_burst_zero_delay
-- 	Core Model for Sync Burst Sram
--
-- 	Revision History:
--   04/23/02  1.0  1) Created VHDL Core.VHD from Verilog Core.V
--   06/05/02  1.1  1) added new signals, DELAY and tKQX.  These signals will
--                     be used to setup the Clock to Data Invalid  spec.
--   07/17/02  1.2  1) Fixed the JTAG State machine
--                  2) changed the SR register to shift out the MSB and shift
--                     in the LSB
--   09/25/02  1.3  1) Removed all nPE pin features
--                  2) Max number of Core addresses is now dynamic
--                  3) Max width of Core data is now dynamic
--                  4) Removed alll reference of JTAG from core, seperate JTAG
--                     model file: GSI_JTAG
--   01/10/03  1.4  1) Created a Write_Array process to remove race conditions
--                  2) Created a Read_Array proccess to remove race conditions
--   02/20/03  1.5  1) Added We and Waddr to Read_Array sensitivity list.
--                  2) Changed the Read_Array process to look at the last
--                     write's byte write setting and determine where to pull the
--                     read data from, either coherency(byte write on) or the
--                     array(byte write off).
--                  3) Added signal Iscd to fix SCD to the right state for NBT
-- 04/03/03   1.6   1) Added a write clock W_k to trigger the Write_Array function
-- 07/09/03   1.7   1) changed NBT write clock to clock off of we2.
--                  2) Delayed the internal clock by 1ns to control the write
--                  3) changed ce to take into account NBT mode
-- 07/23/03   1.8   1) Changed W_K to ignore the byte writes
-- 08/12/03   1.9   1) updated state machine to include seperate read and write
--                     burst states
--                  2) Changed internal bytewrite signal to ignore nW
-- 10/29/03   2.0   1) updated the state machine, changed reference to suspend
--                     to deselect. 
--                  2) added timing functions to core
-- 03/25/04   2.1   1) Updated state machine.  Added deselect and suspend states
--                  2) Fixed other issues with the state machine
-- 04/28/04   2.2   1) Rearranged state that determins Deselect, Burst and Suspend
--
-- 11/01/05   3.0   1) Created BurstRAM only Model
-- 06/21/06   3.1   1) Added Qswitch to control when the IOs turn on or off
--                  2) Delayed the Qxi inteernal data busses instead of the DQx
--                  external Data busses.
--                  3) Added CLK_i2 to control the setting of Qswitch
--                  4) All these changes removed Negative time issue for some simulations
-- 07/18/06   3.2   1) Initialized ce and re to 0 so that Qswitch is not
--                     undfined which can cause bus contention on startup.
-- 10/30/06   3.3   1) Removed enable control of ADSP and ADSC signals in core
-- 05/15/07   3.4   1) updated QSWITCH to include nG as an input
-- 07/13/07   3.5   1) State machine update for ADSP controlled writes / bursting
-- 03/06/08   3.6   1) Added wes signal to control when w_k clock is active.
-- 02/11/09   3.7   1) Removed GSI library requirement
-- 05/06/09   3.8   1) Fixed Deselect state not shifting correctly.
-- 07/20/09   3.9   1) Model now has a reduced memory core.  The size is controlled by A_BITS
--
--
LIBRARY ieee;
USE ieee.std_logic_1164.all;
ENTITY VHDL_BURST_CORE IS
  GENERIC (
    CONSTANT bank_size : integer ;
    CONSTANT A_size    : integer ;
    CONSTANT A_bits    : integer ;
    CONSTANT DQ_size   : integer);
  PORT (
    SIGNAL A     : IN std_logic_vector(A_size - 1 DOWNTO 0);-- address
    SIGNAL DQa   : INOUT std_logic_vector(DQ_size DOWNTO 1);-- byte A data
    SIGNAL DQb   : INOUT std_logic_vector(DQ_size DOWNTO 1);-- byte B data
    SIGNAL DQc   : INOUT std_logic_vector(DQ_size DOWNTO 1);-- byte C data
    SIGNAL DQd   : INOUT std_logic_vector(DQ_size DOWNTO 1);-- byte D data
    SIGNAL DQe   : inout std_logic_vector(DQ_size DOWNTO 1);-- byte E data
    SIGNAL DQf   : inout std_logic_vector(DQ_size DOWNTO 1);-- byte F data
    SIGNAL DQg   : inout std_logic_vector(DQ_size DOWNTO 1);-- byte G data
    SIGNAL DQh   : inout std_logic_vector(DQ_size DOWNTO 1);-- byte H data
    SIGNAL nBa   : IN std_logic;-- bank A write enable
    SIGNAL nBb   : IN std_logic;-- bank B write enable
    SIGNAL nBc   : IN std_logic;-- bank C write enable
    SIGNAL nBd   : IN std_logic;-- bank D write enable
    SIGNAL nBe   : IN std_logic;-- bank E write enable
    SIGNAL nBf   : IN std_logic;-- bank F write enable
    SIGNAL nBg   : IN std_logic;-- bank G write enable
    SIGNAL nBh   : IN std_logic;-- bank H write enable
    SIGNAL CK    : IN std_logic;-- clock
    SIGNAL nBW   : IN std_logic;-- byte write enable
    SIGNAL nGW   : IN std_logic;-- Global write enable
    SIGNAL nE1   : IN std_logic;-- chip enable 1
    SIGNAL E2    : IN std_logic;-- chip enable 2
    SIGNAL nE3   : IN std_logic;-- chip enable 3
    SIGNAL nG    : IN std_logic;-- output enable
    SIGNAL nADV  : IN std_logic;-- Advance not / load
    SIGNAL nADSC : IN std_logic;
    SIGNAL nADSP : IN std_logic;
    SIGNAL ZZ    : IN std_logic;-- power down
    SIGNAL nFT   : IN std_logic;-- Pipeline / Flow through
    SIGNAL nLBO  : IN std_logic;-- Linear Burst Order
    SIGNAL SCD   : IN std_logic;
    SIGNAL HighZ : std_logic_vector(DQ_size downto 1);
    SIGNAL tKQ   : time;
    SIGNAL tKQX  : time);
END VHDL_BURST_CORE;

LIBRARY Std;
ARCHITECTURE GSI_BURST_CORE OF VHDL_BURST_CORE IS
  USE Std.textio.ALL;
  use ieee.std_logic_unsigned.all, ieee.std_logic_arith.all;
  TYPE MEMORY  IS ARRAY (0 TO bank_size) OF std_logic_vector((8*DQ_size)-1 DOWNTO 0);
  TYPE AMEMORY IS ARRAY (0 TO bank_size) OF std_logic_vector(A_size-1 DOWNTO 0);
--   ******** Define Sram Operation Mode    **********************
  shared variable memory_core  : MEMORY ;
  shared variable address_core : AMEMORY ;
-- ---------------------------------------------------------------
--   	Gated SRAM Clock
-- ---------------------------------------------------------------
  SIGNAL clk_i : std_logic;
  SIGNAL clk_i2 : std_logic;
-- ---------------------------------------------------------------
--   	reduced memory stuff
-- ---------------------------------------------------------------
  signal address_used : std_logic_vector(a_bits-1 downto 0) := (others=>'0');
  constant MAX_BITS   : std_logic_vector(a_bits-1 downto 0) := (others=>'1');
-- ---------------------------------------------------------------
--   	Combinatorial Logic
-- ---------------------------------------------------------------
  SIGNAL ADV   : std_logic;-- internal address advance
  SIGNAL W_k   : std_logic;
  SIGNAL R_k   : std_logic;
  SIGNAL BW    : std_logic_vector(7 DOWNTO 0);-- internal byte write enable
  SIGNAL bwi   : std_logic_vector(7  DOWNTO 0);
  SIGNAL Qxi   : std_logic_vector(8*DQ_size-1 DOWNTO 0);-- read data
  SIGNAL Dxi   : std_logic_vector(8*DQ_size-1 DOWNTO 0);-- write data
  SIGNAL addr0 : std_logic_vector(A_size - 1 DOWNTO 0);-- saved address
  SIGNAL addr1 : std_logic_vector(A_size - 1 DOWNTO 0);-- address buffer 1
  SIGNAL baddr : std_logic_vector(A_size - 1 DOWNTO 0);-- burst memory address
  SIGNAL waddr : std_logic_vector(A_size - 1 DOWNTO 0);-- write memory address
  SIGNAL raddr : std_logic_vector(A_size - 1 DOWNTO 0);-- read memory address
  SIGNAL bcnt  : std_logic_vector(1 DOWNTO 0)  := "00";-- burst counter
  SIGNAL we0   : std_logic := '0';
  SIGNAL re0   : std_logic := '0';
  SIGNAL re1   : std_logic := '0';
  SIGNAL re2   : std_logic := '0';
  SIGNAL ce0   : std_logic := '0';
  SIGNAL ce1   : std_logic := '0';
  SIGNAL ce    : std_logic := '0';
  SIGNAL re    : std_logic := '0';
  SIGNAL oe    : std_logic;
  SIGNAL we    : std_logic;
  SIGNAL wes   : std_logic;
  SIGNAL Qswitch: std_logic ;
  SIGNAL state : string (9 DOWNTO 1) := "IDLE     ";

  SIGNAL Check_Time : time   := 1 ns;
  SIGNAL DELAY      : time   := 1 ns;
  SIGNAL OFFSET     : time   := 0 ps;
  SIGNAL GUARD      : boolean:= TRUE;

--  TIMING FUNCTIONS
  function POSEDGE (SIGNAL s : std_ulogic) return BOOLEAN IS
  begin
    RETURN (s'EVENT AND ((To_X01(s'LAST_VALUE) = '0') OR (s = '1')));
  end; 
  function NEGEDGE (SIGNAL s : std_ulogic) return BOOLEAN IS
  begin
    RETURN (s'EVENT AND ((To_X01(s'LAST_VALUE) = '1') OR (s = '0')) );
  end; 
-- END TIMING FUNCTIONS
  
  PROCEDURE shiftnow (SIGNAL addr1 : INOUT std_logic_vector(A_size - 1 DOWNTO 0);
                      SIGNAL re2   : INOUT std_logic;
                      SIGNAL re1   : INOUT std_logic;
                      SIGNAL ce1   : INOUT std_logic;
                      SIGNAL Dxi   : INOUT std_logic_vector(71 DOWNTO 0))
  IS
  BEGIN
    addr1 <= baddr;
    re2 <= re1;
    re1 <= re0;
    ce1 <= ce0;
    Dxi <= (DQh & DQg & DQf & DQe & DQd & DQc & DQb & DQa);
  END;

BEGIN
  PROCESS
  BEGIN
    WAIT UNTIL POSEDGE(CK);
    clk_i  <= NOT ZZ after OFFSET;
    clk_i2 <= NOT ZZ after (OFFSET + 100 ps);
    WAIT UNTIL NEGEDGE(CK);
    clk_i  <= '0' after OFFSET;
    clk_i2 <= '0' after (OFFSET + 100 ps);
  END PROCESS;

-- ---------------------------------------------------------------
--   	State Machine
-- ---------------------------------------------------------------
  st : PROCESS(CK)
    variable tstate : string(9 DOWNTO 1) :="DESELECT ";
    variable twe0 : std_logic := '0';
    variable tre0 : std_logic := '0';
    variable tce0 : std_logic := '0';
  BEGIN
--WAIT UNTIL POSEDGE(CK);
    if POSEDGE(CK) then
      wes <= '1';
    CASE state IS
      WHEN "DESELECT " =>
        if (nE1='0' AND E2='1' AND nE3='0') then
--Checking for ADSC Control
          if (nADSC = '0' and nADSP='1') then
            shiftnow(addr1, re2, re1, ce1, Dxi);
            tre0 := nBW and nGW;
            twe0 := (not nGW or not nBW);
            tce0 := '1';
            addr0 <= A;
            bwi <= BW;
            bcnt <= "00";
            tstate := "NEWCYCLE ";
          end if;
--  Checking for ADSP Control
          if (nADSP = '0') then
            shiftnow(addr1, re2, re1, ce1, Dxi);
            tre0 := '1';
            twe0 := '0';
            tce0 := '1';
            addr0 <= A;
            bcnt <= "00";
            tstate := "LATEWRITE";
          end if;
        END IF;
-- Checking for Deselect
        if (((nE1='1' or E2='0' or nE3= '1') and nADSC = '0') or (nADSP='0' and (E2 = '0' or nE3 = '1')) or (nADSC = '1' and nADSP = '1')) then
          shiftnow(addr1, re2, re1, ce1, Dxi);
          tstate := "DESELECT ";
          twe0 := '0';
          tre0 := '0';
          tce0 := '0';
        END IF;
-- **************************************************
      WHEN "NEWCYCLE " | "BURSTWR  " | "BURSTRD  " | "SUSPBR   " | "LATEWRITE" =>
--Checking for ADSC Control
        if (nADSC = '0' and nADSP='1' and nE1='0' AND E2='1' AND nE3='0') then
          shiftnow(addr1, re2, re1, ce1, Dxi);
          tre0 := (nBW and nGW);
          twe0 :=  (not nGW or not nBW);
          tce0 := '1';
          addr0 <= A;
          bwi <= BW;
          bcnt <= "00";
          tstate := "NEWCYCLE ";
        end if;
--  Checking for ADSP Control
        if (nADSP = '0' and nE1='0' AND E2='1' AND nE3='0') then
          shiftnow(addr1, re2, re1, ce1, Dxi);
          twe0 := '0';
          tre0 := '1';
          tce0 := '1';
          addr0 <= A;
          bcnt <= "00";
          tstate := "LATEWRITE";
        end if;
-- Checking for Deselect
        if (((nE1='1' or E2='0' or nE3='1') and nADSC = '0') or (nADSP = '0' and (E2 = '0' or nE3 = '1'))) then
          shiftnow(addr1, re2, re1, ce1, Dxi);
          tstate := "DESELECT ";
          twe0 := '0';
          tre0 := '0';
          tce0 := '0';
        end if;
--  Checking for Burst Start
        if (nADSC = '1' and nADSP = '1' AND nADV = '0') THEN
          shiftnow(addr1, re2, re1, ce1, Dxi);
          if (nGW = '0' or nBW = '0') then
            tstate := "BURSTWR  ";
            twe0 := '1';
            tre0 := '0';
            bwi <= BW;
          end if;
          if (nGW='1' and nBW = '1') then
            tstate := "BURSTRD  ";
            twe0 := '0';
            tre0 := '1';
          end if;
          tce0 := '1';
          bcnt <= bcnt + "01";
        end if;
--  Checking for a Suspended Burst
        if (nADSC = '1' and nADSP = '1' and nADV = '1') THEN
          shiftnow(addr1, re2, re1, ce1, Dxi);
          tstate := "SUSPBR   ";
          if (we0 = '1' or nGW = '0' or nBW = '0') then
            if (state = "LATEWRITE" and (nGW = '0' or nBW = '0')) then
              wes <= '1';
            else
              wes <= '0';
            end if;
            twe0 := '1';
            tre0 := '0';
            re1 <= '0';
            bwi <= BW;
          elsif (nGW='1' and nBW= '1') then
            twe0 := '0';
            tre0 := '1';
          end if;
          tce0 := '1';
        end if;
      WHEN OTHERS =>
        shiftnow(addr1, re2, re1, ce1, Dxi);
        tstate := "DESELECT ";
        twe0 := '0';
        tre0 := '0';
        tce0 := '0';
        bcnt <= "00";
    END CASE;
    end if;
    state <= tstate;
    we0 <= twe0;
    re0 <= tre0;
    ce0 <= tce0;
  END PROCESS;
-- ---------------------------------------------------------------
--               Data IO Logic
-- ---------------------------------------------------------------
  Write_Array: process (W_k)
    variable addr_t         : integer;
    variable index          : integer := 0;
    variable address_match  : integer := 0;
    variable address_index  : integer := 0;
  begin  -- process Write_Array
    IF (POSEDGE(W_k)) THEN
      if conv_integer(address_used) >0 then
        for index in 0 to conv_integer(address_used)-1 loop
          if address_match = 0 then
            if (address_core(index) = waddr and address_match=0) then
              address_match := 1;
              address_index := index;
            end if;
          end if;
        end loop;  -- index
      end if;
      if address_match = 1 then
        address_core(address_index) := waddr;
        addr_t := address_index;
        address_match := 0;
      else
        if address_used = MAX_BITS then
          report "A Memory overflow has occurred.  Increase Size of A_BITS to increase memory size" severity failure;
        end if;
        address_core(conv_integer(address_used)) := waddr;
        addr_t := conv_integer(address_used);
        address_used <= address_used + 1;
      end if;
      IF bw(0) = '1' THEN
        memory_core(conv_integer(addr_t))((DQ_size-1) downto 0)   := Dxi((DQ_size-1) downto 0);
      END IF;
      IF bw(1) = '1' THEN
        memory_core(conv_integer(addr_t))(2*(DQ_size)-1 downto DQ_size)  := Dxi(2*(DQ_size)-1 downto (DQ_size));
      END IF;
      IF bw(2) = '1' THEN
        memory_core(conv_integer(addr_t))(3*(DQ_size)-1 downto 2*(DQ_size)) := Dxi(3*(DQ_size)-1 downto 2*(DQ_size));
      END IF;
      IF bw(3) = '1' THEN
        memory_core(conv_integer(addr_t))(4*(DQ_size)-1 downto 3*(DQ_size)) := Dxi(4*(DQ_size)-1 downto 3*(DQ_size));
      END IF;
      IF bw(4) = '1' THEN
        memory_core(conv_integer(addr_t))(5*(DQ_size)-1 downto 4*(DQ_size)) := Dxi(5*(DQ_size)-1 downto 4*(DQ_size));
      END IF;
      IF bw(5) = '1' THEN
        memory_core(conv_integer(addr_t))(6*(DQ_size)-1 downto 5*(DQ_size)) := Dxi(6*(DQ_size)-1 downto 5*(DQ_size));
      END IF;
      IF bw(6) = '1'  THEN
        memory_core(conv_integer(addr_t))(7*(DQ_size)-1 downto 6*(DQ_size)) := Dxi(7*(DQ_size)-1 downto 6*(DQ_size));
      END IF;
      IF bw(7) = '1'  THEN
        memory_core(conv_integer(addr_t))(8*(DQ_size)-1 downto 7*(DQ_size)) := Dxi(8*(DQ_size)-1 downto 7*(DQ_size));
      END IF;
   END IF;
  end process Write_Array;

  Read_Array: process (r_k)
    variable index          : integer := 0;
    variable address_match  : integer := 0;
    variable address_index  : integer := 0;
  begin  -- process Read_Array
    if conv_integer(address_used) > 0 then
      for index in 0 to conv_integer(address_used) - 1 loop
        if address_match = 0 then
          if address_core(index) = raddr then
            address_match := 1;
            address_index := index;
          end if;
        end if;
      end loop;  -- indev
    end if;
    if address_match = 1 then
      address_match := 0;
      Qxi <= transport memory_core(address_index) after DELAY - 200 ps;
    else
      Qxi <= transport (others=>'X') after DELAY - 200 ps;
    end if;
end process Read_Array;

-- check it -t option is active and set correctly
time_ck : process (CLK_i2)
begin
  check_time <= CK'last_event;
  assert check_time /= 0ns report "Resolution needs to be set to 100ps for modelSIM use vsim -t 100ps <>" severity FAILURE;
end process time_ck;

ADS_SET : process (CK)
begin
  if posedge(ck) then
    ADV   <=  not nADV;
    BW(0) <=  not nGW or (NOT nBa and not nBW);
    BW(1) <=  not nGW or (NOT nBb and not nBW);
    BW(2) <=  not nGW or (NOT nBc and not nBW);
    BW(3) <=  not nGW or (NOT nBd and not nBW);
    BW(4) <=  not nGW or (NOT nBe and not nBW);
    BW(5) <=  not nGW or (NOT nBf and not nBW);
    BW(6) <=  not nGW or (NOT nBg and not nBW);
    BW(7) <=  not nGW or (NOT nBh and not nBW);
  end if;
end process ADS_SET;

q_switch : process (CLK_i2)
begin                                 --read clock controls outputs
  Qswitch <= transport (not nG and(re and ce)) after DELAY - 200 ps;
end process q_switch;


W_k   <=(wes and we and not ck) after (OFFSET);
R_k   <= re1 and clk_i after 100 ps when nFT='1' else re0 and clk_i after 100 ps;
we    <= we0;
re    <= re1 when nFT='1' else re0;
ce    <= ce1 when (SCD='0' AND re2 = '1') else ce0;
oe    <= re AND ce;

baddr <= (addr0(A_size - 1 DOWNTO 2)&(bcnt(1) XOR addr0(1))&(bcnt(0) XOR addr0(0))) when nLBO='1'
           else (addr0(A_size - 1 DOWNTO 2)&(addr0(1 DOWNTO 0)+bcnt));

waddr <=  addr0 when ADV='0' else baddr;
raddr <=  addr1 when nFT='1' else baddr;

DELAY <=  tKQX when (nG='1' OR not ((we='1' and re='1') or oe='1') OR ZZ='1') else tKQ;

DQa   <= GUARDED Qxi((DQ_size)-1 downto 0)   when Qswitch = '1' else (others=>'Z');
DQb   <= GUARDED Qxi(2*(DQ_size)-1 downto (DQ_size))  when Qswitch = '1' else (others=>'Z');
DQc   <= GUARDED Qxi(3*(DQ_size)-1 downto 2*(DQ_size)) when Qswitch = '1' else (others=>'Z');
DQd   <= GUARDED Qxi(4*(DQ_size)-1 downto 3*(DQ_size)) when Qswitch = '1' else (others=>'Z');
DQe   <= GUARDED Qxi(5*(DQ_size)-1 downto 4*(DQ_size)) when Qswitch = '1' else (others=>'Z');
DQf   <= GUARDED Qxi(6*(DQ_size)-1 downto 5*(DQ_size)) when Qswitch = '1' else (others=>'Z');
DQg   <= GUARDED Qxi(7*(DQ_size)-1 downto 6*(DQ_size)) when Qswitch = '1' else (others=>'Z');
DQh   <= GUARDED Qxi(8*(DQ_size)-1 downto 7*(DQ_size)) when Qswitch = '1' else (others=>'Z');
END GSI_BURST_CORE;
